/*
 * Copyright 2013-2020 Signal Messenger, LLC
 * SPDX-License-Identifier: AGPL-3.0-only
 */
package org.whispersystems.textsecuregcm.storage;

import static com.codahale.metrics.MetricRegistry.name;

import com.codahale.metrics.MetricRegistry;
import com.codahale.metrics.SharedMetricRegistries;
import com.codahale.metrics.Timer;
import com.codahale.metrics.Timer.Context;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import org.jdbi.v3.core.transaction.TransactionIsolationLevel;
import org.whispersystems.textsecuregcm.storage.mappers.AccountRowMapper;
import org.whispersystems.textsecuregcm.util.Constants;
import org.whispersystems.textsecuregcm.util.SystemMapper;

public class Accounts implements AccountStore {

  public static final String ID = "id";
  public static final String UID = "uuid";
  public static final String NUMBER = "number";
  public static final String DATA = "data";
  public static final String VERSION = "version";

  private static final ObjectMapper mapper = SystemMapper.getMapper();

  private final MetricRegistry metricRegistry        = SharedMetricRegistries.getOrCreate(Constants.METRICS_NAME);
  private final Timer          createTimer           = metricRegistry.timer(name(Accounts.class, "create"          ));
  private final Timer          updateTimer           = metricRegistry.timer(name(Accounts.class, "update"          ));
  private final Timer          getByNumberTimer      = metricRegistry.timer(name(Accounts.class, "getByNumber"     ));
  private final Timer          getByUuidTimer        = metricRegistry.timer(name(Accounts.class, "getByUuid"       ));
  private final Timer          getAllFromTimer       = metricRegistry.timer(name(Accounts.class, "getAllFrom"      ));
  private final Timer          getAllFromOffsetTimer = metricRegistry.timer(name(Accounts.class, "getAllFromOffset"));
  private final Timer          deleteTimer           = metricRegistry.timer(name(Accounts.class, "delete"          ));
  private final Timer          vacuumTimer           = metricRegistry.timer(name(Accounts.class, "vacuum"          ));

  private final FaultTolerantDatabase database;

  public Accounts(FaultTolerantDatabase database) {
    this.database = database;
    this.database.getDatabase().registerRowMapper(new AccountRowMapper());
  }

  @Override
  public boolean create(Account account) {
    return database.with(jdbi -> jdbi.inTransaction(TransactionIsolationLevel.SERIALIZABLE, handle -> {
      try (Timer.Context ignored = createTimer.time()) {
        final Map<String, Object> resultMap = handle.createQuery("INSERT INTO accounts (" + NUMBER + ", " + UID + ", " + DATA + ") VALUES (:number, :uuid, CAST(:data AS json)) ON CONFLICT(number) DO UPDATE SET " + DATA + " = EXCLUDED.data, " + VERSION + " = accounts.version + 1 RETURNING uuid, version")
                                                    .bind("number", account.getNumber())
                                                    .bind("uuid", account.getUuid())
                                                    .bind("data", mapper.writeValueAsString(account))
                                                    .mapToMap()
                                                    .findOnly();

        final UUID uuid = (UUID) resultMap.get(UID);
        final int version = (int) resultMap.get(VERSION);

        boolean isNew = uuid.equals(account.getUuid());
        account.setUuid(uuid);
        account.setVersion(version);
        return isNew;
      } catch (JsonProcessingException e) {
        throw new IllegalArgumentException(e);
      }
    }));
  }

  @Override
  public void update(Account account) throws ContestedOptimisticLockException {
    database.use(jdbi -> jdbi.useHandle(handle -> {
      try (Timer.Context ignored = updateTimer.time()) {
        final int newVersion = account.getVersion() + 1;
        int rowsModified = handle.createUpdate("UPDATE accounts SET " + DATA + " = CAST(:data AS json), " + VERSION + " = :newVersion WHERE " + UID + " = :uuid AND " + VERSION + " = :version")
              .bind("uuid", account.getUuid())
              .bind("data", mapper.writeValueAsString(account))
              .bind("version", account.getVersion())
              .bind("newVersion", newVersion)
              .execute();

        if (rowsModified == 0) {
          throw new ContestedOptimisticLockException();
        }

        account.setVersion(newVersion);

      } catch (JsonProcessingException e) {
        throw new IllegalArgumentException(e);
      }
    }));
  }

  @Override
  public Optional<Account> get(String number) {
    return database.with(jdbi -> jdbi.withHandle(handle -> {
      try (Timer.Context ignored = getByNumberTimer.time()) {
        return handle.createQuery("SELECT * FROM accounts WHERE " + NUMBER + " = :number")
                     .bind("number", number)
                     .mapTo(Account.class)
                     .findFirst();
      }
    }));
  }

  @Override
  public Optional<Account> get(UUID uuid) {
    return database.with(jdbi -> jdbi.withHandle(handle -> {
      try (Timer.Context ignored = getByUuidTimer.time()) {
        return handle.createQuery("SELECT * FROM accounts WHERE " + UID + " = :uuid")
                     .bind("uuid", uuid)
                     .mapTo(Account.class)
                     .findFirst();
      }
    }));
  }

  public AccountCrawlChunk getAllFrom(UUID from, int length) {
    final List<Account> accounts = database.with(jdbi -> jdbi.withHandle(handle -> {
      try (Context ignored = getAllFromOffsetTimer.time()) {
        return handle.createQuery("SELECT * FROM accounts WHERE " + UID + " > :from ORDER BY " + UID + " LIMIT :limit")
            .bind("from", from)
            .bind("limit", length)
            .mapTo(Account.class)
            .list();
      }
    }));
    return buildChunkForAccounts(accounts);
  }

  public AccountCrawlChunk getAllFrom(int length) {
    final List<Account> accounts = database.with(jdbi -> jdbi.withHandle(handle -> {
      try (Timer.Context ignored = getAllFromTimer.time()) {
        return handle.createQuery("SELECT * FROM accounts ORDER BY " + UID + " LIMIT :limit")
                     .bind("limit", length)
                     .mapTo(Account.class)
                     .list();
      }
    }));

    return buildChunkForAccounts(accounts);
  }

  private AccountCrawlChunk buildChunkForAccounts(final List<Account> accounts) {
    return new AccountCrawlChunk(accounts, accounts.isEmpty() ? null : accounts.get(accounts.size() - 1).getUuid());
  }

  @Override
  public void delete(final UUID uuid) {
    database.use(jdbi -> jdbi.useHandle(handle -> {
      try (Timer.Context ignored = deleteTimer.time()) {
        handle.createUpdate("DELETE FROM accounts WHERE " + UID + " = :uuid")
                .bind("uuid", uuid)
                .execute();
      }
    }));
  }

  public void vacuum() {
    database.use(jdbi -> jdbi.useHandle(handle -> {
      try (Timer.Context ignored = vacuumTimer.time()) {
        handle.execute("VACUUM accounts");
      }
    }));
  }

}
